#   Convoy is a WSGI app for loading multiple files in the same request.
#   Copyright (C) 2010-2012  Canonical, Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import json
import re
import sys
import glob
import optparse


DETAILS_RE = re.compile(
    "YUI\.add\([\'\"\s]*([^\'\"]*)[\'\"\s]*,.*?function.*?"
    "[\'\"\s]*[0-9\.]*[\'\"\s]*"
    "({[^{}a-zA-Z]*(use|requires|optional|after|"
    "supersedes|omit|skinnable)[^a-zA-Z][\s\'\":]*[^{}]*})\s*\);", re.M | re.S)
DETAILS_FIND = re.compile(
    '''(?<!")([a-z]+)(?:\s+)?:''',
    re.M | re.S)
DETAILS_REPLACE = r'"\1":'

LITERAL_RE = re.compile("([\[ ]+)\"([\w\.\+-]+)\"([^:])")
NAME_RE = re.compile("[\.\+-]")


def extract_metadata(src):
    """Extract metadata about an YUI module, given it's source."""
    metadata = []
    for entry in DETAILS_RE.finditer(src):
        name, details, ignore = entry.groups()
        details = details.replace('\'', '"')
        details = DETAILS_FIND.sub(DETAILS_REPLACE, details)
        details = json.loads(details)
        details["name"] = name
        metadata.append(details)
    return metadata


class Builder:

    def __init__(self, name, src_dir, output=None,
                 prefix="", exclude_regex=None, ext=True, include_skin=True):
        """Create a new Builder.

        :param name: The name of the global variable name that will
            contain the modules configuration.
        :param src_dir: The directory containing the source files.
        :param output: The output filename for the module metadata.
        :param prefix: A prefix to be added to the relative path.
        :param exclude_regex: A regex that will exclude file
            paths from the final rollup.  -min and -debug versions
            will still be built.
        :param ext: Default value for the 'ext' setting. Set to
            'False' to use modules with a local combo loader.
        :param include_skin: If False, the generated metadata won't include
            skin modules and all javascript modules will have skinnable=False.
            Defaults to True.
        """
        self.name = name
        self.output = output
        self.prefix = prefix
        self.src_dir = src_dir
        self.exclude_regex = exclude_regex
        self.include_skin = include_skin
        self.ext = ext

    def log(self, msg):
        sys.stdout.write(msg + '\n')

    def fail(self, msg):
        """An error was encountered, abort build."""
        sys.stderr.write(msg + '\n')
        sys.exit(1)

    def file_is_excluded(self, filepath):
        """Is the given file path excluded from the rollup process?"""
        if not self.exclude_regex:
            # Include everything.
            return False

        return re.search(self.exclude_regex, filepath)

    def generate_metadata(self, fnames, root, var_name, out):
        """Extract metadata from a group of files and write it out."""
        metadata = []

        for fname in fnames:
            self.log("Extracting metadata from '%s'" % fname)
            data = open(fname, "r").read()
            meta = extract_metadata(data)
            prefix = ""
            if self.prefix and not prefix.endswith("/"):
                prefix = self.prefix + "/"
            for entry in meta:
                # According to the source of the YUI loader module:
                #
                #   The default path for the YUI library is the
                #   minified version of the files (e.g., event-min.js).
                #
                # To make it easier for everyone, let's use the same
                # convention here, and use the minified path.
                relpath = (
                    prefix + fname.replace(root + os.path.sep, "")
                    ).replace(os.path.sep, "/")
                entry["path"] = "%s-min%s" % (
                    os.path.splitext(relpath))

                entry["type"] = "js"
                entry["ext"] = self.ext
                if self.include_skin and entry.get("skinnable"):
                    self.generate_skin_modules(entry, metadata, root)
            metadata.extend(meta)

        modules = {}

        # Include config keys in literals.
        all_literals = ["use", "requires", "optional", "after", "type",
                        "supersedes", "ext", "path", "css", "js", "skinnable"]
        for meta in metadata:
            name = meta.pop("name")
            modules[name] = meta
            all_literals.append(name)
            all_literals.extend(meta.get("use", ()))
            all_literals.extend(meta.get("requires", ()))
            all_literals.extend(meta.get("after", ()))
            all_literals.append(meta["type"])

        # For each string literal we are interested in, generate a
        # variable declaration for the string literal, to improve
        # minification.
        #
        # The variable name is generated by replacing [".", "+", "-"]
        # with an underscore and then make that the variable name,
        # uppercase.
        #
        # We'll save a mapping of literal -> variable name here for
        # reuse below on the re.sub() helper function.
        literals_map = dict([(literal, NAME_RE.sub("_", literal).upper())
                             for literal in set(all_literals)])

        # This re.sub() helper function operates on the JSON-ified
        # representation of the modules, by looking for string
        # literals that occur over the JSON structure but *not* as
        # attribute names.
        #
        # If a string literal is found that matches the list of
        # literals we have declared as variables above, then replace
        # the it by the equivalent variable, otherwise return the
        # original string.
        def literal_sub(match):
            literal = match.group(2)
            if literal in literals_map:
                return match.group(1) + literals_map[literal] + match.group(3)
            return match.group(0)


        linebreak = ",\n  "
        variables_decl = "var SKIN_SAM_PREFIX = 'skin-sam-'" + linebreak
        if self.prefix:
             variables_decl += "PREFIX = '%s'%s" % (self.prefix, linebreak)
        extra_variables = []
        for literal, variable in sorted(literals_map.iteritems()):
            extra_variable = "%s = %s" % (
                variable, ('"%s"' % literal).replace(
                    '"skin-sam-', 'SKIN_SAM_PREFIX + "'))
            if self.prefix:
                extra_variable = extra_variable.replace(
                    '"%s.' % self.prefix, 'PREFIX + ".')
            extra_variables.append(extra_variable)
        variables_decl += linebreak.join(extra_variables)

        # Default 'after' modules from YUI Loader. Might have to
        # be changed in the future, if YUI itself changes.
        core_css = ["cssreset", "cssfonts",
                    "cssgrids", "cssreset-context",
                    "cssfonts-context",
                    "cssgrids-context"]

        # Append a few more helper variables for our own use.
        variables_decl += ",\n  ".join(
            ["",
             "modules = {}",
             "TRUE = true",
             "FALSE = false",
             "CORE_CSS = %s" % LITERAL_RE.sub(literal_sub,
                                              json.dumps(core_css)),
             "module_info",
             "after_list"])

        modules_decl = []
        for module_name, module_info in sorted(modules.iteritems()):
            module_decl = [
                "modules[%s] = module_info = {}" %
                NAME_RE.sub("_", module_name).upper()]
            for key, value in sorted(module_info.iteritems()):
                if value is True or value is False:
                    value = str(value).upper()
                elif value in ("css", "js"):
                    value = value.upper()
                else:
                    value = LITERAL_RE.sub(literal_sub, json.dumps(value))
                if key == "after" and module_info["type"] == "css":
                    # It's easy to think that doing 'CORE_CSS +  %(values)s'
                    # instead of using concat would work, but it doesn't;
                    # you'll end up with a string instead of a list.
                    module_decl.append("after_list = CORE_CSS");
                    module_decl.append("after_list.concat(%s)" % value);
                    value = "after_list"
                if key == "path":
                    value = value.replace(
                        '"%s/' % self.prefix, 'PREFIX + "/')
                module_decl.append("module_info[%s] = %s" %
                                   (key.upper(), value))
            modules_decl.append(";\n  ".join(module_decl))

        modules_decl = ";\n\n  ".join(modules_decl)

        module_config = open(out, "w")
        try:
            module_config.write("""var %s = (function(){
  %s;

  %s;

  return modules;
})();""" % (var_name, variables_decl, modules_decl))
        finally:
            module_config.close()

    def generate_skin_modules(self, entry, metadata, root):
        # Generate a skin module definition, since YUI assumes that
        # the path starts with the module name, and that breaks with
        # our structure.
        #
        # Follow lazr-js conventions and look for any file in the skin
        # assets directory.
        module_names = []
        by_name = {}

        prefix = ""
        if self.prefix and not prefix.endswith("/"):
            prefix = self.prefix + "/"

        if entry.get("requires"):
            # If the base module requires other modules, extend
            # the after entry with the (expected) skins for those
            # modules to force our skin to be loaded after those.
            after = ["skin-sam-%s" % s
                     for s in entry["requires"]]

        assets = os.path.join(
            os.path.dirname(entry["path"][len(prefix):]), "assets")
        sam = os.path.join(assets, "skins", "sam")
        css_assets = glob.glob(os.path.join(root, sam, "*-skin.css"))

        for fname in css_assets:
            if not os.path.exists(fname):
                # If the file doesn't exist, don't create a module to
                # load it.
                continue

            # Compute a module name for this asset.
            module_name = os.path.basename(fname)[:-len("-skin.css")]
            skin_module_name = "skin-sam-%s" % entry["name"]
            # If the computed module_name does not match the
            # Javascript module name without the namespace, then use
            # it as a postfix to disambiguate possibly multiple
            # modules.
            package = entry["name"].split(".")[-1]
            if module_name != package and len(css_assets) > 1:
                skin_module_name = "%s+%s" % (skin_module_name, module_name)

            css = (fname.replace(root + os.path.sep, "")
                   ).replace(os.path.sep, "/")
            module = {"name": skin_module_name,
                      "after": after[:],
                      "type": "css",
                      "ext": self.ext,
                      "path": prefix + css}
            by_name[skin_module_name] = module
            module_names.append(skin_module_name)
            metadata.append(module)

        # All assets under the skin have been looked at. Now look for
        # a "-core.css" asset, following lazr-js conventions and add
        # it as a requirement for the previously-found assets.
        for module_name in module_names:
            name = os.path.basename(
                by_name[module_name]["path"])[:-len("-skin.css")]
            fname = os.path.join(root, assets, "%s-core.css" % name)
            if not os.path.exists(fname):
                # No core CSS asset exists for this skin, skip
                # generating a module for it.
                continue

            skin_module_name = "%s+core" % module_name
            css = (fname.replace(root + os.path.sep, "")
                   ).replace(os.path.sep, "/")
            module = {"name": skin_module_name,
                      "after": after[:],
                      "type": "css",
                      "ext": self.ext,
                      "path": prefix + css}

            requires = by_name[module_name].setdefault("requires", [])
            requires.append(skin_module_name)
            after = by_name[module_name].setdefault("after", [])
            after.append(skin_module_name)
            metadata.append(module)

    def do_build(self):
        included_files = []

        for root, dirnames, fnames in os.walk(self.src_dir):
            for fname in glob.glob(os.path.join(root, '*.js')):
                if not self.file_is_excluded(fname):
                    included_files.append(fname)

        self.generate_metadata(included_files, self.src_dir,
                               self.name, self.output)


def get_options():
    """Parse the command line options."""
    parser = optparse.OptionParser(
        usage="%prog [options]",
        description=(
            "Create YUI module metadata from extension modules. "
            ))
    parser.add_option(
        '-n', '--name', dest='name', default='LAZR_MODULES',
        help=('The global variable name used to hold the modules config.'))
    parser.add_option(
        '-o', '--output', dest='output',
        help=('The output filename for the module metadata.'))
    parser.add_option(
        '-p', '--prefix', dest='prefix', default="",
        help=('A prefix to be added to the relative path.'))
    parser.add_option(
        '-s', '--srcdir', dest='src_dir',
        help=('The directory containing the src files.'))
    parser.add_option(
        '-e', '--ext', dest='ext', default=False,
        action="store_true",
        help=('Default value for the "ext" configuration setting.'))
    parser.add_option(
        '-x', '--exclude', dest='exclude_regex',
        default=None, metavar='REGEX',
        help=('Exclude any files that match the given regular expressions.'))
    parser.add_option(
        '-k', '--no-skin', dest='no_skin', default=False, action="store_true",
        help=('Do not include skin files in the list of modules and set '
              'skinnable=False for all modules.'))
    return parser.parse_args()


def main():
   options, args = get_options()
   if options.src_dir is None:
       options.src_dir = os.getcwd()
   Builder(
       name=options.name,
       src_dir=os.path.abspath(options.src_dir),
       output=options.output,
       prefix=options.prefix,
       exclude_regex=options.exclude_regex,
       ext=options.ext,
       include_skin=not options.no_skin,
       ).do_build()



#   Convoy is a WSGI app for loading multiple files in the same request.
#   Copyright (C) 2010-2012  Canonical, Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import cgi
import os
import re
import urlparse


CHUNK_SIZE = 2 << 12
URL_RE = re.compile("url\([ \"\']*([^ \"\']+)[ \"\']*\)")
URL_PARSE = re.compile("/([^/]*).*?$")


def relative_path(from_file, to_file):
    """Return the relative path between from_file and to_file."""
    dir_from, base_from = os.path.split(from_file)
    dir_to, base_to = os.path.split(to_file)
    path = os.path.relpath(dir_to, dir_from)
    if path == ".":
        return base_to
    return os.path.join(path, base_to)


def parse_url(url):
    """Parse a combo URL.

    Returns the list of arguments in the original order.
    """
    scheme, loc, path, query, frag = urlparse.urlsplit(url)
    return parse_qs(query)


def parse_qs(query):
    """Parse a query string.

    Returns the list of arguments in the original order.
    """
    params = cgi.parse_qsl(query, keep_blank_values=True)
    return tuple([param for param, value in params])


def combine_files(fnames, root, resource_prefix="", rewrite_urls=True):
    """Combine many files into one.

    Returns an iterator with the combined content of all the
    files. The relative path to root will be included as a comment
    between each file.
    """

    resource_prefix = resource_prefix.rstrip("/")
    for fname in fnames:
        file_ext = os.path.splitext(fname)[-1]
        basename = os.path.basename(fname)
        full = os.path.abspath(os.path.join(root, fname))
        yield "/* " + fname + " */\n"
        if not full.startswith(root) or not os.path.exists(full):
            yield "/* [missing] */\n"
        else:
            with open(full, "r") as f:
                if file_ext == ".css" and rewrite_urls:
                    file_content = f.read()
                    src_dir = os.path.dirname(full)
                    relative_parts = relative_path(
                        os.path.join(root, basename), src_dir).split(
                        os.path.sep)

                    def fix_relative_url(match):
                        url = match.group(1)
                        # Don't modify absolute URLs or 'data:' urls.
                        if (url.startswith("http") or
                            url.startswith("/") or
                            url.startswith("data:")):
                            return match.group(0)
                        parts = relative_parts + url.split("/")
                        result = []
                        for part in parts:
                            if part == ".." and result and result[-1] != "..":
                                result.pop(-1)
                                continue
                            result.append(part)
                        return "url(%s)" % "/".join(
                            filter(None, [resource_prefix] + result))
                    file_content = URL_RE.sub(fix_relative_url, file_content)
                    yield file_content
                    yield "\n"
                else:
                    while True:
                        chunk = f.read(CHUNK_SIZE)
                        if not chunk:
                            yield "\n"
                            break
                        yield chunk


def combo_app(root, resource_prefix="", rewrite_urls=True):
    """A simple YUI Combo Service WSGI app.

    Serves any files under C{root}, setting an appropriate
    C{Content-Type} header.
    """
    root = os.path.abspath(root)

    def app(environ, start_response, root=root):
        # Path hint uses the rest of the url to map to files on disk based off
        # the root specified to convoy.
        path_hint = environ['PATH_INFO'].strip('/')
        fnames = parse_qs(environ["QUERY_STRING"])
        content_type = "text/plain"
        if fnames:
            if fnames[0].endswith(".js"):
                content_type = "text/javascript"
            elif fnames[0].endswith(".css"):
                content_type = "text/css"
        else:
            start_response("404 Not Found", [("Content-Type", content_type)])
            return ("Not Found",)

        # Take any prefix in the url route into consideration for the root to
        # find files.
        updated_root = os.path.join(root, path_hint)
        # Enforce that the updated root is not outside the original root.
        absroot = os.path.abspath(updated_root)
        if not absroot.startswith(os.path.abspath(root)):
            start_response("400 Bad Request", [("Content-Type", content_type)])
            return ("Bad Request",)
        else:
            start_response("200 OK", [("Content-Type", content_type)])

        return combine_files(fnames, updated_root, resource_prefix,
                             rewrite_urls=rewrite_urls)
    return app



#   Convoy is a WSGI app for loading multiple files in the same request.
#   Copyright (C) 2010-2012  Canonical, Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import difflib
import textwrap
from unittest import defaultTestLoader

import mocker
from paste.fixture import TestApp

from convoy.combo import combo_app
from convoy.combo import combine_files
from convoy.combo import parse_url


class ComboTestBase(object):

    def makeSampleFile(self, root, fname, content):
        content = textwrap.dedent(content).strip()
        full = os.path.join(root, fname)
        parent = os.path.dirname(full)
        if not os.path.exists(parent):
            os.makedirs(parent)
        return self.makeFile(content=content, path=full)

    def assertTextEquals(self, expected, got):
        expected = textwrap.dedent(expected).strip()
        got = textwrap.dedent(got).strip()
        diff = difflib.unified_diff(expected.splitlines(),
                                    got.splitlines(), lineterm="")
        self.assertEquals(expected, got, "\n" + "\n".join(diff))


class ComboTest(ComboTestBase, mocker.MockerTestCase):

    def test_parse_url_keeps_order(self):
        """Parsing a combo loader URL returns an ordered list of filenames."""
        self.assertEquals(
            parse_url(("http://yui.yahooapis.com/combo?"
                       "3.0.0/build/yui/yui-min.js&"
                       "3.0.0/build/oop/oop-min.js&"
                       "3.0.0/build/event-custom/event-custom-min.js&")),
            ("3.0.0/build/yui/yui-min.js",
             "3.0.0/build/oop/oop-min.js",
             "3.0.0/build/event-custom/event-custom-min.js"))

    def test_combine_files_includes_filename(self):
        """Combining files should include their relative filename."""
        test_dir = self.makeDir()

        self.makeSampleFile(
            test_dir,
            os.path.join("yui", "yui-min.js"),
            "** yui-min **"),
        self.makeSampleFile(
            test_dir,
            os.path.join("oop", "oop-min.js"),
            "** oop-min **"),
        self.makeSampleFile(
            test_dir,
            os.path.join("event-custom", "event-custom-min.js"),
            "** event-custom-min **"),

        expected = "\n".join(("/* yui/yui-min.js */",
                              "** yui-min **",
                              "/* oop/oop-min.js */",
                              "** oop-min **",
                              "/* event-custom/event-custom-min.js */",
                              "** event-custom-min **"))
        self.assertEquals(
            "".join(combine_files(["yui/yui-min.js",
                                   "oop/oop-min.js",
                                   "event-custom/event-custom-min.js"],
                                  root=test_dir)).strip(),
            expected)

    def test_combine_css_makes_relative_path(self):
        """
        Combining CSS files makes URLs in CSS declarations relative to the
        target path.
        """
        test_dir = self.makeDir()

        self.makeSampleFile(
            test_dir,
            os.path.join("widget", "assets", "skins", "sam", "widget.css"),
            """\
            /* widget skin */
            .yui-widget {
               background: url("img/bg.png");
            }
            """)

        self.makeSampleFile(
            test_dir,
            os.path.join("editor", "assets", "skins", "sam", "editor.css"),
            """\
            /* editor skin */
            .yui-editor {
               background: url("img/bg.png");
            }
            """)

        expected = """
            /* widget/assets/skins/sam/widget.css */
            /* widget skin */
            .yui-widget {
               background: url(widget/assets/skins/sam/img/bg.png);
            }
            /* editor/assets/skins/sam/editor.css */
            /* editor skin */
            .yui-editor {
               background: url(editor/assets/skins/sam/img/bg.png);
            }
            """
        self.assertTextEquals(
            "".join(combine_files(["widget/assets/skins/sam/widget.css",
                                   "editor/assets/skins/sam/editor.css"],
                                  root=test_dir)).strip(),
            expected)

    def test_combine_css_leaves_absolute_urls_untouched(self):
        """
        Combining CSS files does not touch absolute URLs in
        declarations.
        """
        test_dir = self.makeDir()

        self.makeSampleFile(
            test_dir,
            os.path.join("widget", "assets", "skins", "sam", "widget.css"),
            """
            /* widget skin */
            .yui-widget {
               background: url("/static/img/bg.png");
            }
            """)

        self.makeSampleFile(
            test_dir,
            os.path.join("editor", "assets", "skins", "sam", "editor.css"),
            """
            /* editor skin */
            .yui-editor {
               background: url("http://foo/static/img/bg.png");
            }
            """)

        expected = """
            /* widget/assets/skins/sam/widget.css */
            /* widget skin */
            .yui-widget {
               background: url("/static/img/bg.png");
            }
            /* editor/assets/skins/sam/editor.css */
            /* editor skin */
            .yui-editor {
               background: url("http://foo/static/img/bg.png");
            }
            """
        self.assertTextEquals(
            "".join(combine_files(["widget/assets/skins/sam/widget.css",
                                   "editor/assets/skins/sam/editor.css"],
                                  root=test_dir)).strip(),
            expected)

    def test_combine_css_leaves_data_uris_untouched(self):
        """
        Combining CSS files does not touch data uris in
        declarations.
        """
        test_dir = self.makeDir()

        self.makeSampleFile(
            test_dir,
            os.path.join("widget", "assets", "skins", "sam", "widget.css"),
            """\
            /* widget skin */
            .yui-widget {
               background: url("data:image/gif;base64,base64-data");
            }
            """)

        self.makeSampleFile(
            test_dir,
            os.path.join("editor", "assets", "skins", "sam", "editor.css"),
            """\
            /* editor skin */
            .yui-editor {
               background: url("data:image/gif;base64,base64-data");
            }
            """)

        expected = """
            /* widget/assets/skins/sam/widget.css */
            /* widget skin */
            .yui-widget {
               background: url("data:image/gif;base64,base64-data");
            }
            /* editor/assets/skins/sam/editor.css */
            /* editor skin */
            .yui-editor {
               background: url("data:image/gif;base64,base64-data");
            }
            """
        self.assertTextEquals(
            "".join(combine_files(["widget/assets/skins/sam/widget.css",
                                   "editor/assets/skins/sam/editor.css"],
                                  root=test_dir)).strip(),
            expected)

    def test_missing_file_is_ignored(self):
        """If a missing file is requested we should still combine others."""
        test_dir = self.makeDir()

        self.makeSampleFile(
            test_dir,
            os.path.join("yui", "yui-min.js"),
            "** yui-min **"),
        self.makeSampleFile(
            test_dir,
            os.path.join("event-custom", "event-custom-min.js"),
            "** event-custom-min **"),

        expected = "\n".join(("/* yui/yui-min.js */",
                              "** yui-min **",
                              "/* oop/oop-min.js */",
                              "/* [missing] */",
                              "/* event-custom/event-custom-min.js */",
                              "** event-custom-min **"))
        self.assertEquals(
            "".join(combine_files(["yui/yui-min.js",
                                   "oop/oop-min.js",
                                   "event-custom/event-custom-min.js"],
                                  root=test_dir)).strip(),
            expected)

    def test_no_parent_hack(self):
        """If someone tries to hack going up the root, he'll get a miss."""
        test_dir = self.makeDir()

        self.makeSampleFile(
            test_dir,
            os.path.join("oop", "oop-min.js"),
            "** oop-min **"),

        root = os.path.join(test_dir, "root", "lazr")
        os.makedirs(root)

        hack = "../../oop/oop-min.js"
        self.assertTrue(os.path.exists(os.path.join(root, hack)))

        expected = "\n".join(("/* ../../oop/oop-min.js */",
                              "/* [missing] */"))
        self.assertEquals(
            "".join(combine_files([hack], root=root)).strip(),
            expected)

    def test_no_absolute_path_hack(self):
        """
        If someone tries to fetch an absolute file, it'll be as if a missing
        file was requested.
        """
        test_dir = self.makeDir()

        hack = "/etc/passwd"
        self.assertTrue(os.path.exists("/etc/passwd"))

        expected = "/* /etc/passwd */\n/* [missing] */"
        self.assertEquals(
            "".join(combine_files([hack], root=test_dir)).strip(),
            expected)

    def test_no_traversing_out_of_root(self):
        """
        If someone were to use .. in the path make sure we don't let them out
        of the root of the combo loader
        """
        test_dir = self.makeDir()

        hack = ".."

        expected = "/* .. */\n/* [missing] */"
        self.assertEquals(
            "".join(combine_files([hack], root=test_dir)).strip(),
            expected)

        # from /tmp/somedir we want to try to walk up to / and into
        # etc/password
        hack = "../../etc/password"
        expected = "/* ../../etc/password */\n/* [missing] */"
        self.assertEquals(
            "".join(combine_files([hack], root=test_dir)).strip(),
            expected)

    def test_combine_css_adds_custom_prefix(self):
        """
        Combining CSS files makes URLs in CSS declarations relative to the
        target path. It's also possible to specify an additional prefix for the
        rewritten URLs.
        """
        test_dir = self.makeDir()

        self.makeSampleFile(
            test_dir,
            os.path.join("widget", "assets", "skins", "sam", "widget.css"),
            """
            /* widget skin */
            .yui-widget {
               background: url("img/bg.png");
            }
            """)
        self.makeSampleFile(
            test_dir,
            os.path.join("editor", "assets", "skins", "sam", "editor.css"),
            """
            /* editor skin */
            .yui-editor {
               background: url("img/bg.png");
            }
            """)

        expected = """
            /* widget/assets/skins/sam/widget.css */
            /* widget skin */
            .yui-widget {
               background: url(/static/widget/assets/skins/sam/img/bg.png);
            }
            /* editor/assets/skins/sam/editor.css */
            /* editor skin */
            .yui-editor {
               background: url(/static/editor/assets/skins/sam/img/bg.png);
            }
            """
        self.assertTextEquals(
            "".join(combine_files(["widget/assets/skins/sam/widget.css",
                                   "editor/assets/skins/sam/editor.css"],
                                  root=test_dir,
                                  resource_prefix="/static/")).strip(),
            expected)

    def test_rewrite_url_normalizes_parent_references(self):
        """URL references in CSS files get normalized for parent dirs."""
        test_dir = self.makeDir()
        files = [
            self.makeSampleFile(
                test_dir,
                os.path.join("yui", "base", "base.css"),
                ".foo{background-image:url(../../img.png)}"),
            ]

        expected = """
            /* yui/base/base.css */
            .foo{background-image:url(img.png)}
            """
        self.assertTextEquals(
            "".join(combine_files(["yui/base/base.css"],
                                  root=test_dir)).strip(),
            expected)


class WSGIComboTest(ComboTestBase, mocker.MockerTestCase):

    def setUp(self):
        self.root = self.makeDir()
        self.app = TestApp(combo_app(self.root))

    def test_combo_app_sets_content_type_for_js(self):
        """The WSGI App should set a proper Content-Type for Javascript."""
        self.makeSampleFile(
            self.root,
            os.path.join("yui", "yui-min.js"),
            "** yui-min **"),
        self.makeSampleFile(
            self.root,
            os.path.join("oop", "oop-min.js"),
            "** oop-min **"),
        self.makeSampleFile(
            self.root,
            os.path.join("event-custom", "event-custom-min.js"),
            "** event-custom-min **"),

        expected = "\n".join(("/* yui/yui-min.js */",
                              "** yui-min **",
                              "/* oop/oop-min.js */",
                              "** oop-min **",
                              "/* event-custom/event-custom-min.js */",
                              "** event-custom-min **"))

        res = self.app.get("/?" + "&".join(
            ["yui/yui-min.js",
             "oop/oop-min.js",
             "event-custom/event-custom-min.js"]), status=200)
        self.assertEquals(res.headers, [("Content-Type", "text/javascript")])
        self.assertEquals(res.body.strip(), expected)

    def test_combo_app_sets_content_type_for_css(self):
        """The WSGI App should set a proper Content-Type for CSS."""
        self.makeSampleFile(
            self.root,
            os.path.join("widget", "skin", "sam", "widget.css"),
            "/* widget-skin-sam */"),

        expected = "/* widget/skin/sam/widget.css */\n/* widget-skin-sam */"

        res = self.app.get("/?" + "&".join(
            ["widget/skin/sam/widget.css"]), status=200)
        self.assertEquals(res.headers, [("Content-Type", "text/css")])
        self.assertEquals(res.body.strip(), expected)

    def test_no_filename_gives_404(self):
        """If no filename is included, a 404 should be returned."""
        res = self.app.get("/", status=404)
        self.assertEquals(res.headers, [("Content-Type", "text/plain")])
        self.assertEquals(res.body, "Not Found")

    def test_combo_respects_path_hints(self):
        """If I add path info into the combo url, convoy should use it."""
        # we want the app to be rooted at /tmp and we'll adjust the paths via
        # url
        app = TestApp(combo_app('/tmp'))

        # create paths we'll store files into. All of them end up in
        # /tmp/xxxxx. We want to point the root of the combo loader at /tmp
        # and use url pathing to route to the right set of files for final
        # usage.
        first_root = self.makeDir()
        second_root = self.makeDir()

        # we need to parse where the actual roots are located to know what
        # path to add to the urls
        first_base = os.path.basename(first_root)
        second_base = os.path.basename(second_root)

        self.makeSampleFile(
            first_root,
            os.path.join("yui", "yui-min.js"),
            "/* yui-min */"),

        self.makeSampleFile(
            second_root,
            os.path.join("yui", "yui-min.js"),
            "/* yui-min-2 */"),

        expected = "\n".join(("/* yui/yui-min.js */",
                              "/* yui-min */"))
        expected2 = "\n".join(("/* yui/yui-min.js */",
                               "/* yui-min-2 */"))

        res = app.get("/%s/?%s" % (first_base,
            "yui/yui-min.js"), status=200)
        self.assertEquals(res.body.strip(), expected)

        res = app.get("/%s/?%s" % (second_base, "&".join(
            ["yui/yui-min.js"])), status=200)
        self.assertEquals(res.body.strip(), expected2)

    def test_path_hint_cant_break_root(self):
        """You should not be able to get outside of the root via path hint."""
        app = TestApp(combo_app('/tmp'))

        # we attempt to use the path hint to reassign root to /etc
        # and ask for passwd, it's inside the new adjusted root
        app.get("/../etc?yui-min&passwd", status=400)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)



#   Convoy is a WSGI app for loading multiple files in the same request.
#   Copyright (C) 2010-2012  Canonical, Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
from unittest import defaultTestLoader, TestCase

import mocker

from convoy.meta import Builder, extract_metadata


class TestBuilder(Builder):
    """A special CSSComboFile that doesn't log to the stdout."""

    def log(self, msg):
        pass


class ExtractMetadataTest(TestCase):

    def test_extract_single_module(self):
        """
        Extracting the metadata of a file containing a single module
        should successfully extract it's requirements.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.base', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "base"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.base")
        self.assertEquals(metadata[0]["requires"], ["node", "base"])

    def test_extract_multiple_modules(self):
        """
        Extracting the metadata of a file that has more than one
        module registered should find all the registrations.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.base', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "base"]});
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "anim", "event"]});
        """)

        self.assertEquals(len(metadata), 2)

        self.assertEquals(metadata[0]["name"], "lazr.base")
        self.assertEquals(metadata[0]["requires"], ["node", "base"])

        self.assertEquals(metadata[1]["name"], "lazr.anim")
        self.assertEquals(metadata[1]["requires"], ["node", "anim", "event"])

    def test_extract_multi_line(self):
        """
        If the module registration is split between more than one
        line, we should still be able to extract it.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "anim",
                              "event"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])

    def test_extract_odd_spacing(self):
        """
        Extracting metadata should work just fine if you have odd
        spacing between the requirements.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": [ "node" ,"anim"  ,  "event" ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])

    def test_extract_with_use(self):
        """
        Having both 'use' and 'require in the same declaration should
        not break our parser.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"use": ["dom"], "requires": [ "node" ,"anim"  ,  "event" ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_with_requires_before_use(self):
        """
        The order between 'use' and 'requires' should not matter.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": [ "node" ,"anim"  ,  "event" ], "use": ["dom"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_requires_in_new_line(self):
        """
        Even if the metadata is split between multiple lines, we
        should still be able to extract it.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {
        "requires": [
          "node" ,"anim"  ,  "event"
        ],
        "use": [
          "dom"
        ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_metadata_not_metadata(self):
        """
        Don't get metadata from things that ARE NOT metadata.
        """
        metadata = extract_metadata("""
        YUI.add('test', function(Y) {
            this._dds[h[i]] = new YAHOO.util.DragDrop(this._handles[h[i]], this.get('id') + '-handle-' + h, { useShim: this.get('useShim') });
        }, '1.0', {requires: ['dom']});
        """)

        self.assertEquals(metadata[0]["requires"], ["dom"])

    def test_extract_metadata_single_quotes(self):
        """
        If the javascript is using object literals, we should adjust for
        simplejson.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {
        requires: [
          'node' ,'anim'  ,  'event'
        ],
        use: [
          'dom'
        ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])


    def test_extract_has_no_quotes(self):
        """
        If the javascript is using object literals, we should adjust for
        simplejson.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {
        requires: [
          "node" ,"anim"  ,  "event"
        ],
        use: [
          "dom"
        ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_has_single_quotes(self):
        """
        If the javascript is using single quotes, we should adjust for
        simplejson.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {
        'requires': [
          "node" ,"anim"  ,  "event"
        ],
        use: [
          "dom"
        ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_has_use(self):
        """
        If the javascript is has the word 'cause' the regex shouldn't match
        use in it.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {
        'requires': [
          "node" ,"anim"  ,  "event"
        ],
        use: [
          "cause",
          "dom"
        ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["cause", "dom"])

    def test_extract_with_use_and_no_requires(self):
        """
        No failure should happen if we have 'use' but no 'requires'.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"use": ["dom"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_with_all_options(self):
        """
        A very complex registration, with all the options. We should
        be able to parse all of it.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
           X = Y.bind(function(Y){
               Y.log('goodbye world');
           }, Y);
        }, '0.1', {"use": ["dom"], "omit": ["nono"],
                   "optional": ["meh"],
                   "supersedes": ["old-anim"],
                   "after": ["lazr.base"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["use"], ["dom"])
        self.assertEquals(metadata[0]["omit"], ["nono"])
        self.assertEquals(metadata[0]["optional"], ["meh"])
        self.assertEquals(metadata[0]["supersedes"], ["old-anim"])
        self.assertEquals(metadata[0]["after"], ["lazr.base"])

class GenerateMetadataTest(mocker.MockerTestCase):

    def test_generate_metadata_simple(self):
        """
        Verify that the builder can extract and generates a metadata
        file in the expected format.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"]});
        """, basename="anim.js", dirname=anim)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="")
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        modules = "\n\n".join(got.split("\n\n")[1:-1])
        self.assertEquals(prefix, "var LAZR_CONFIG = ")
        self.assertTrue('[PATH] = "anim/anim-min.js"' in modules)
        self.assertFalse('[SKINNABLE] = FALSE' in modules)
        self.assertTrue('[TYPE] = JS' in modules)
        self.assertTrue('[USE] = [DOM]' in modules)
        self.assertTrue('[EXT] = TRUE' in modules)

    def test_extract_skinnable(self):
        """
        If a module is skinnable, we'll generate the module
        registration for it's skin ourselves, to prevent YUI from
        auto-generating one, since it assumes the skin path starts
        with the module name.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"], "requires": ["widget"], "skinnable": true});
        """, basename="anim.js", dirname=anim)

        skin = self.makeDir(path=os.path.join(anim, "assets", "skins", "sam"))
        self.makeFile("""\
        /* anim-skin.css */
        """, basename="anim-skin.css", dirname=skin)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="",
                              prefix="lazr")
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        modules = "\n\n".join(got.split("\n\n")[1:-1])
        self.assertEquals(prefix, "var LAZR_CONFIG = ")
        self.assertTrue(
            '[PATH] = PREFIX + '
            '"/anim/assets/skins/sam/anim-skin.css"' in modules)
        self.assertTrue(
            '[PATH] = PREFIX + "/anim/anim-min.js"' in modules)
        self.assertTrue('[SKINNABLE] = TRUE' in modules)
        self.assertTrue('[TYPE] = JS' in modules)
        self.assertTrue('[TYPE] = CSS' in modules)
        self.assertTrue('[USE] = [DOM]' in modules)
        self.assertTrue('[REQUIRES] = [WIDGET]' in modules)
        self.assertTrue(('after_list = CORE_CSS') in modules)
        self.assertTrue(('after_list.concat([SKIN_SAM_WIDGET])') in modules)
        self.assertTrue(('[AFTER] = after_list') in modules)
        self.assertTrue('[EXT] = TRUE' in modules)

    def test_extract_no_prefix_doesnt_add_prefix(self):
        """
        If a prefix is not specified, it won't be prepended to the filename.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"], "requires": ["widget"], "skinnable": true});
        """, basename="anim.js", dirname=anim)

        skin = self.makeDir(path=os.path.join(anim, "assets", "skins", "sam"))
        self.makeFile("""\
        /* anim-skin.css */
        """, basename="anim-skin.css", dirname=skin)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="",
                              prefix=None)
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        modules = "\n\n".join(got.split("\n\n")[1:-1])
        self.assertEquals(prefix, "var LAZR_CONFIG = ")
        self.assertFalse('  PREFIX =' in got, got)
        self.assertTrue(
            '[PATH] = "anim/assets/skins/sam/anim-skin.css"' in modules)
        self.assertTrue(
            '[PATH] = "anim/anim-min.js"' in modules)
        self.assertTrue('[SKINNABLE] = TRUE' in modules)
        self.assertTrue('[TYPE] = JS' in modules)
        self.assertTrue('[TYPE] = CSS' in modules)
        self.assertTrue('[USE] = [DOM]' in modules)
        self.assertTrue('[REQUIRES] = [WIDGET]' in modules)
        self.assertTrue(('after_list = CORE_CSS') in modules)
        self.assertTrue(('after_list.concat([SKIN_SAM_WIDGET])') in modules)
        self.assertTrue(('[AFTER] = after_list') in modules)
        self.assertTrue('[EXT] = TRUE' in modules)

    def test_extract_no_prefix_no_directory(self):
        """
        If a prefix is not specified, and the filename is directly at the root,
        things don't blow up.
        """
        root = self.makeDir()
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"], "requires": ["widget"], "skinnable": true});
        """, basename="anim.js", dirname=root)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="",
                              prefix=None)
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        modules = "\n\n".join(got.split("\n\n")[1:-1])
        self.assertEquals(prefix, "var LAZR_CONFIG = ")
        self.assertFalse('  PREFIX =' in got, got)
        self.assertTrue(
            '[PATH] = "anim-min.js"' in modules)
        self.assertTrue('[TYPE] = JS' in modules)
        self.assertTrue('[USE] = [DOM]' in modules)
        self.assertTrue('[REQUIRES] = [WIDGET]' in modules)
        self.assertTrue('[EXT] = TRUE' in modules)

    def test_extract_skinnable_with_lazr_conventions(self):
        """
        LazrJS conventionally uses a "-core.css" file to load core CSS rules,
        and a "-skin.css" file to store skinnable rules.

        If we find this filesystem structure, let's generate different
        CSS modules and hook them up accordingly. The filename doesn't
        need to match the module name.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"], "requires": ["widget"], "skinnable": true});
        """, basename="anim.js", dirname=anim)

        assets = self.makeDir(path=os.path.join(anim, "assets"))
        self.makeFile("""\
        /* purty-anim-core.css */
        """, basename="purty-anim-core.css", dirname=assets)

        skin = self.makeDir(path=os.path.join(assets, "skins", "sam"))
        self.makeFile("""\
        /* purty-anim-skin.css */
        """, basename="purty-anim-skin.css", dirname=skin)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="",
                              prefix="lazr")
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        blocks = got.split("\n\n")
        modules = "\n\n".join(blocks[1:-1])
        self.assertEquals(prefix, "var LAZR_CONFIG = ")

        self.assertTrue(
            '[PATH] = PREFIX + "/anim/assets/purty-anim-core.css"' in modules)
        self.assertTrue(
            '[PATH] = PREFIX + "/anim/assets/skins/sam/purty-anim-skin.css"'
            in modules)
        self.assertTrue(
            '[PATH] = PREFIX + "/anim/anim-min.js"' in modules)
        self.assertTrue('[SKINNABLE] = TRUE' in modules)
        self.assertTrue('[TYPE] = JS' in modules)
        self.assertTrue('[TYPE] = CSS' in modules)
        self.assertTrue('[USE] = [DOM]' in modules)
        self.assertTrue('[REQUIRES] = [WIDGET]' in modules)
        self.assertTrue('[REQUIRES] = [SKIN_SAM_LAZR_ANIM_CORE]' in modules)
        self.assertTrue(('after_list = CORE_CSS') in modules)
        self.assertTrue(
            ('after_list.concat([SKIN_SAM_WIDGET, SKIN_SAM_LAZR_ANIM_CORE])')
            in modules)
        self.assertTrue(('[AFTER] = after_list') in modules)
        self.assertTrue('[EXT] = TRUE' in modules)

    def test_literal_variables_available(self):
        """
        All literals should be specified as var strings.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {
            "after": [],
            "ext": false,
            "optional": ["midget"],
            "path": "/path/to/module",
            "requires": ["widget"],
            "skinnable": true,
            "supersedes": [],
            "use": ["dom"]
        });
        """, basename="anim.js", dirname=anim)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="",
                              prefix="lazr")
        builder.do_build()

        got = open(output, "r").read()
        self.assertIn('AFTER = "after",', got)
        self.assertIn('EXT = "ext",', got)
        self.assertIn('OPTIONAL = "optional",', got)
        self.assertIn('PATH = "path",', got)
        self.assertIn('REQUIRES = "requires",', got)
        self.assertIn('SKINNABLE = "skinnable",', got)
        self.assertIn('SUPERSEDES = "supersedes",', got)
        self.assertIn('USE = "use",', got)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)



#   Convoy is a WSGI app for loading multiple files in the same request.
#   Copyright (C) 2010-2012  Canonical, Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
from distutils.core import setup

# If setuptools is present, use it to find_packages(), and also
# declare our dependency on epsilon.
extra_setup_args = {}
try:
    from setuptools import find_packages
    extra_setup_args = {
        'install_requires': ['Paste'],
        'tests_require': ['nose', 'mocker']
        }
except ImportError:
    def find_packages(exclude=None):
        """
        Compatibility wrapper.

        Taken from storm setup.py.
        """
        packages = []
        for directory, subdirectories, files in os.walk("convoy"):
            if '__init__.py' in files:
                packages.append(directory.replace(os.sep, '.'))
        return packages

setup(
    name="convoy",
    version="0.2.4",
    description="A combo WSGI application for use with YUI",
    author="Canonical Javascripters",
    url="https://launchpad.net/convoy",
    license="AGPL",
    packages=find_packages(exclude=('convoy.tests',)),
    include_package_data=True,
    zip_safe=False,
    test_suite="nose.collector",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
       ],
    **extra_setup_args
    )

